import os
import sysconfig
import tomllib
from pathlib import Path


def find(
    name: str,
    return_path_not_found: Path,
    return_parent: bool = True,
) -> Path:
    current: Path = Path(__file__).parent

    while current != current.parent:
        if current.joinpath(name).exists():
            if return_parent:
                return current
            return current.joinpath(name)

        current = current.parent

    return return_path_not_found


thispath = Path(__file__).parent

PROJECT_ROOT = find("pyproject.toml", return_parent=True, return_path_not_found=thispath)
PROJECT_TOML = find("pyproject.toml", return_parent=False, return_path_not_found=PROJECT_ROOT)
VERSION_PATH = find("__version__.py", return_parent=False, return_path_not_found=PROJECT_ROOT)
CONFIG_PATH = find("settings.toml", return_parent=False, return_path_not_found=PROJECT_ROOT)
APP_NAME_PATH = find("__app_name__.py", return_parent=False, return_path_not_found=PROJECT_ROOT)
LOGGING_CONFIGURATION_PATH = find(
    "logging_configuration.json",
    return_parent=False,
    return_path_not_found=PROJECT_ROOT,
)
LOGS_PATH = find("logs", return_parent=False, return_path_not_found=PROJECT_ROOT)

USER_SCRIPTS_PATH = sysconfig.get_path("scripts", f"{os.name}_user")

if PROJECT_ROOT is None:
    exit

assert PROJECT_TOML

if Path(PROJECT_TOML).exists():
    with PROJECT_TOML.open("rb") as toml_file:
        PYPROJECT = tomllib.load(toml_file)



from dynaconf import Dynaconf

from src.hledger_tools.info import CONFIG_PATH, PYPROJECT


def settings_generator() -> Dynaconf:
    return Dynaconf(
        envvar_prefix=PYPROJECT["project"]["name"],
        environments=True,
        settings_files=["settings.toml", ".secrets.toml"],
        load_dotenv=True,
        root_path=CONFIG_PATH,
    )



from hledger_tools.resources import Entry, Loan, Mark, Posting


def make_loan_entries(loan: Loan, cash: list[str], expense: list[str], liability: list[str]) -> list[Entry]:
    payments = loan.payment_schedule().values()

    entries: list[Entry] = list()
    for payment in payments:
        postings: list = list()

        postings.append(
            Posting(
                account=cash,
                amount=payment.amount * -1,
                comment="",
            )
        )
        postings.append(
            Posting(
                account=expense,
                amount=payment.interest,
                comment="",
            )
        )
        postings.append(
            Posting(
                account=liability,
                amount=payment.principal,
                comment="",
            )
        )

        entries.append(
            Entry(
                date=payment.term,
                status=Mark.UNMARKED,
                code="",
                description=loan.name,
                comment="",
                postings=postings,
            )
        )

    return entries



from datetime import date
from decimal import Decimal
from pathlib import Path

import click
from moneyed import Money

from hledger_tools.make_loan_entries import make_loan_entries
from hledger_tools.resources import Entry, Loan, Mark, Posting

DEFAULT_CURRENCY = "USD"


@click.group
def cli():
    pass


@cli.command
@click.argument("output")
def loan_journal_entries(output: str):
    # need to get a bunch of information from the user
    loan_name = click.prompt(text="Loan Description", type=click.STRING)
    loan_begin_args = click.prompt(text="Beginning Date (YYYY-MM-DD)", type=click.STRING).split("-")
    loan_begin = date(int(loan_begin_args[0]), int(loan_begin_args[1]), int(loan_begin_args[2]))
    term_months = click.prompt(text="Loan Term in Years", type=click.INT) * 12
    annual_interest = round(Decimal(click.prompt(text="Annual Interest Rate (ex: 0.02)", type=click.FLOAT)), 3)
    principal = Money(click.prompt(text="Loan Amount", type=click.FLOAT), DEFAULT_CURRENCY)

    cash = click.prompt(text="Cash Account", type=click.STRING).split(":")
    expense = click.prompt(text="Interest Expense Account", type=click.STRING).split(":")
    liability = click.prompt(text="Liability Account", type=click.STRING).split(":")

    loan = Loan(
        name=loan_name,
        begin=loan_begin,
        term_months=term_months,
        annual_interest=annual_interest,
        principal=principal,
    )

    postings = [
        Posting(account=cash, amount=loan.principal, comment=""),
        Posting(account=liability, amount=loan.principal * -1, comment=""),
    ]
    initial_entry = Entry(
        date=loan_begin,
        code="",
        status=Mark.UNMARKED,
        description=f"Opening Loan: {loan.principal}",
        comment="Opening Loan",
        postings=postings,
    )

    entries = [initial_entry]
    entries.extend(make_loan_entries(loan=loan, cash=cash, expense=expense, liability=liability))
    results = [entry.to_str for entry in entries]

    click.echo()
    click.echo(loan)
    click.echo()

    for entry in entries[:3]:
        click.echo(entry.to_str)
        click.echo()
        click.echo()

    click.echo()
    click.echo(".")
    click.echo(".")
    click.echo(".")
    click.echo()

    for entry in entries[-3:]:
        click.echo(entry.to_str)
        click.echo()
        click.echo()

    path = Path(output)
    path.write_text("\n\n\n".join(results))



from dataclasses import dataclass
from datetime import date
from decimal import Decimal
from enum import Enum

from moneyed import Money

DATE_FORMAT = "%Y-%m-%d"


class Mark(Enum):
    UNMARKED = ""
    PENDING = "!"
    CLEARED = "*"


@dataclass
class Posting:
    account: list[str]
    amount: Money
    comment: str


@dataclass
class Entry:
    date: date
    status: Mark
    code: str
    description: str
    comment: str
    postings: list[Posting]

    @property
    def to_str(self):
        result: str = str()

        result = f"{self.date.strftime(format=DATE_FORMAT)} {self.description}"

        if self.comment:
            result += f" ; {self.comment}"

        for post in self.postings:
            temp = str(post.amount).replace("-$", "$-")
            result += f"\n   {':'.join(post.account)}  {temp}"

            if post.comment:
                result += f" ; {post.comment}"

        return result


def add_month(current_date: date) -> date:
    year = current_date.year
    month = current_date.month
    day = current_date.day

    month = month + 1
    if month > 12:
        year += 1
        month = 1

    return date(year, month, day)


@dataclass
class Payment:
    beginning_principal: Money
    term: date
    interest: Money
    principal: Money
    ending_principal: Money

    @property
    def amount(self):
        return self.interest + self.principal

    def __repr__(self):
        return (
            "<Payment("
            f"term={self.term.strftime(format=DATE_FORMAT)}, "
            f"begin={self.beginning_principal}, "
            f"pmt_i={self.interest}, "
            f"pmt_p={self.principal}, "
            f"total={self.amount}, "
            f"end={self.ending_principal}"
            ")>"
        )


@dataclass
class Loan:
    name: str
    begin: date
    term_months: int
    annual_interest: Decimal
    principal: Money

    def __post_init__(self):
        self.interest = self.annual_interest / 12
        one_plus_term = (1 + self.interest) ** (self.term_months)
        self.payment = self.principal * self.interest * one_plus_term / (one_plus_term - 1)

    def payment_schedule(self) -> dict[date, Payment]:
        payments: dict[date, Payment] = dict()
        current: date = self.begin

        while self.principal > Money(0, "USD"):
            interest_pmt = self.principal * self.interest
            principal_pmt = self.payment - interest_pmt

            if principal_pmt > self.principal:
                principal_pmt = self.principal

            beginning_principal = self.principal
            ending_principal = self.principal - principal_pmt

            payments[current] = Payment(
                beginning_principal=beginning_principal.round(2),
                term=current,
                interest=interest_pmt.round(2),
                principal=principal_pmt.round(2),
                ending_principal=ending_principal.round(2),
            )

            self.principal -= principal_pmt
            current = add_month(current)

        return payments
